/***************************************************************
 * Atmega32 SImple OS
 * Walter Schilling, Instructor
 * September, 2011
 *
 * This file contains the implementation for a very simple
 * embedded operating system.  In essence, it provides for a simple,
 * round robin scheduler based upon timer 0.
 * Processes may be stopped and started at will by the user.
 ***************************************************************
 */
#include <stdint.h>
#include <stdbool.h>
#include "inc/hw_ints.h"
#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "driverlib/sysctl.h"
#include "driverlib/interrupt.h"
#include "driverlib/gpio.h"
#include "driverlib/timer.h"
#include "os.h"
#include "portdef.h"
#include "inc/hw_ints.h"
#include "inc/hw_memmap.h"
#include "inc/hw_types.h"


#include "driverlib/debug.h"
#include "driverlib/fpu.h"
#include "driverlib/gpio.h"
#include "driverlib/interrupt.h"
#include "driverlib/pin_map.h"
#include "driverlib/rom.h"
#include "driverlib/sysctl.h"
#include "driverlib/timer.h"
#include "utils/uartstdio.h"

extern void getsp(); /* declare external asm function */
extern void writesp();
extern void save_context();
extern void load_context();
void systick_handler();

uint32_t stackp = 0;              /* define global variable        */

// This structure contains the structure for a process control block.
// It includes the process state, the process id (pid), and a stack pointer.
struct processStruct
{
  int8_t processState;
  uint8_t processID;
  uint32_t stackPtr;
};

/* This table holds the process control blocks for the operating system.
   Each process has a pcb associated with it. */
static struct processStruct pcbs[MAXNUMBEROFPROCESSES];

// This variable keeps track of the number of executing processes.
static uint8_t numberOfExecutingProcesses;

// This variable holds the current process index, which is the index into the pcb block for the currently executing process.
static volatile uint8_t currentProcessIndex; 

// These are function prototypes for the functions used within this module.
static void enableTimer();
static uint8_t scheduleNextProcess(uint8_t currentProcess);
void trapProcessDied();

/*************************************************************** 
 * This method will start the operation of the OS.  The process 
 * control blocks will be initialized, and the calling process 
 * will be given process block 0.
 *
 ***************************************************************/
void startOS()
{
  // Initialize the pcbs to their default state.
  uint8_t index;

  for (index = 0; index < MAXNUMBEROFPROCESSES; index++)
  {
    pcbs[index].processState = UNINITIALZED;
	pcbs[index].processID = 0;
	pcbs[index].stackPtr = NULL;
  }

  // Now that the initialization is done, set up pcbs[0] to represent the main process.
  pcbs[0].processState = RUNNING;
  pcbs[0].processID = 1; 
  pcbs[0].stackPtr = (void*)STACKSTART;
  numberOfExecutingProcesses = 1;

  // Set the current process index.
  currentProcessIndex = 0;

  // Enable the multitasking interrupt. 
  enableTimer();

  // Enable interrupts
  tBoolean temp = IntMasterEnable();
};


/*************************************************************** 
 * addProcess
 * This method will add a new process to the OS.  The process
 * is passed in as a void pointer.
 *
 * Parameters:
 *  function (void*) This is the function which is to be the new process.
 * Return
 * The methdo will return the assigned process id (pid) or
 * -1 if the process can not be added. 
 *
 ***************************************************************/
int8_t addProcess(void* function)
{
  // Read in the state of sreg and store it.
  tBoolean temp;
  int8_t pcb_index;
  uint8_t index;
  int8_t retVal;

  // Disable interrupts.
  temp = IntMasterDisable();

  // Determine the next stack to use.
  pcb_index = 0;
  retVal = -1;

  while ((pcb_index < MAXNUMBEROFPROCESSES) && (retVal == -1))
  {
    // Find an emptypcb for this process.
	if (pcbs[pcb_index].processState == UNINITIALZED)
	{
	  // We've found a spot.
	  retVal = 0;
	}
	else
	{
	  pcb_index++;
	}
  }
  
  // If we have an empty pcb, then we can start this process.
  if (retVal != -1)
  {
    // Setup the pcb for this new process.
    pcbs[pcb_index].processState = NEW;
    pcbs[pcb_index].processID = pcb_index+1;
    pcbs[pcb_index].stackPtr = ((uint32_t)STACKSTART - (STACKSIZE*pcb_index));
	retVal = pcbs[pcb_index].processID;

	// Place the function for the trap which will cause the OS to kill the given process onto the stack.
	uint32_t *tsp = ((uint32_t*)pcbs[pcb_index].stackPtr);
	*tsp = 0x21000000;  // Default PSR value.
	tsp--;

	*tsp = 0x61000000;
	tsp--;

	// Place the return value on the stack.
	*tsp = (uint32_t)function;
	tsp--;

	*tsp = 0xB8B;
	tsp--;

	*tsp = 0x00000000;
	tsp--;

	*tsp = 0x00000001;
	tsp--;

	*tsp = 0x00000E0C;
	tsp--;

	*tsp = 0x00080000;
	tsp--;

	*tsp = 0xE000E100;
	tsp--;
	// Place the return function that we are to go to on the stack.

	*tsp = 0xFFFFFFF9;
	tsp--;

	*tsp = 0x00000001;
	tsp--;

	*tsp = 0x00000E0C;
	tsp--;

	*tsp = 0x00080000;
	tsp--;

  	// Now move the stack pointer location down by 37 bytes, which is the 
  	// number of pushes when the ISR fires.
  	for (index = 0; index < 8; index++)
  	{
  		*tsp = (uint32_t)index;
  		tsp--;
  	}

  	tsp++;

  	pcbs[pcb_index].stackPtr = (uint32_t)tsp;

  	// Increment the number of executing processes.
  	numberOfExecutingProcesses++;
  }
  else
  {
    // We can not start this process because there is no pcb available.

  }
  // Restore sreg to its previous state so that the scheduler task can fire and run this new process.
  //if (temp==true)
  //{
	  IntMasterEnable();
  //}

  // Return the appropriate return value.
  return retVal;
}

/*************************************************************** 
 * trapProcessDied
 * This method will trap, causing the given process to terminate.
 * Essentially, this means that the given process has ended its execution.
 *
 * Parameters:
 *  None
 * Return
 *  None
 ***************************************************************/
void trapProcessDied()
{
  // Read in the state of sreg and store it.
  uint8_t temp;

  // Start by disabling interrupts
  // Disable interrupts.
  temp = IntMasterDisable();

  // Set the given process to be dead.
  pcbs[currentProcessIndex].processState = UNINITIALZED;

  // Restore sreg to its previous state so that the scheduler task can fire and run this new process.
  if (temp==true)
  {
	  IntMasterEnable();
  }

  // Wait for the interrupt to fire.
  while (1==1);
}

/*************************************************************** 
 * terminateProcess
 * This method will terminate an active process.
 *
 * Parameters:
 *  pid (int8_t) This is the pid for the process that is to be terminated.
 * Return
 *  None
 ***************************************************************/
void terminateProcess(int8_t pid)
{
	tBoolean temp;
  int8_t pcb_index;
  bool matchFound;  

  // Disable interrupts.
  temp = IntMasterDisable();

  matchFound = false;
  pcb_index = 0;
  
  while ((matchFound == false) && (pcb_index < MAXNUMBEROFPROCESSES))
  {
    // Determine if the pid matches.
	if (pcbs[pcb_index].processID == pid)
	{
	  // We've found the spot.  Now kill the process.
      pcbs[pcb_index].processState = UNINITIALZED;
      matchFound = true;
	}
	else
	{
	  pcb_index++;
	}
  }
  // Restore interrupts.
  if (temp==true)
  {
	  IntMasterEnable();
  }

  // If we are the currently executing process, trap to the OS to force another
  // process to be scheduled.
  if (pcb_index == currentProcessIndex)
  {
    trapProcessDied();
  }
}

/*************************************************************** 
 * enableTimer
 * This method will enable timer 0 as a periodic timer.  Essentially,
 * when it fires, the scheduler will run to determine the next process 
 * which should execute.
 *
 * Parameters:
 *  None
 * Return
 *  None
 ***************************************************************/
static void enableTimer()
{
	  //
	  // Enable the peripherals used by this example.
	  //
	  SysCtlPeripheralEnable(SYSCTL_PERIPH_TIMER0);

	  //
	  // Enable processor interrupts.
	  //
	  ROM_IntMasterEnable();

	  //
	  // Configure the two 32-bit periodic timers.
	  //
	  TimerConfigure(TIMER0_BASE, TIMER_CFG_PERIODIC);
	  TimerLoadSet(TIMER0_BASE, TIMER_A, SysCtlClockGet()/100);

	  //
	  // Setup the interrupts for the timer timeouts.
	  //
	  IntEnable(INT_TIMER0A);

	  //
	  // Enable the timers.
	  //
	  TimerEnable(TIMER0_BASE, TIMER_A);

	  // Enable the interrupt.
	  TimerIntEnable(TIMER0_BASE, TIMER_TIMA_TIMEOUT);

}

// This method will determine the next process which is to run.
/*************************************************************** 
 * scheduleNextProcess
 * This method will determine the next process that is to execute.
 * This is done through a round robin approach.
 *
 * Parameters:
 *  currentProcess (uint8_t) This is offset in the pcb block for the current process.
 * Return
 *  uint8_t The pcb index offset into the table for the next process which is to execute.
 ***************************************************************/
static uint8_t scheduleNextProcess(uint8_t currentProcess)
{
  uint8_t index;
  uint8_t retVal;
  bool matchFound;

  index = currentProcess+1;
  matchFound = false;  
  retVal = currentProcess;

  while ((index <= (currentProcess + MAXNUMBEROFPROCESSES)) && (matchFound == false))
  {
    if ((pcbs[(index%MAXNUMBEROFPROCESSES)].processState == RUNNING) || (pcbs[(index%MAXNUMBEROFPROCESSES)].processState == NEW))
	{
	  retVal = (index%MAXNUMBEROFPROCESSES);
	  matchFound = true;
	}
	else
	{
      index++;
    }
  }
  return retVal;
}

unsigned long g_ulFlags;

/***************************************************************
 * This ISR will cause the scheduler to run.  In doing so, it will
 * determine the next process that is ready and execute it.
 ***************************************************************/
void systick_handler()
{
    //
    // Clear the timer interrupt.
    //
    ROM_TimerIntClear(TIMER0_BASE, TIMER_TIMA_TIMEOUT);

    //
    // Toggle the flag for the first timer.
    //
    //HWREGBITW(&g_ulFlags, 0) ^= 1;

    // Determine the current process based on the stack pointer.
    // Get the current stack.
    save_context();
    getsp();

    uint32_t stackPointer = stackp;
    uint32_t offset = ((uint32_t)STACKSTART - stackPointer) / STACKSIZE;

    // Store the stack pointer.
    pcbs[offset].stackPtr = stackPointer;

    // Obtain the next process to run.
    offset = scheduleNextProcess(offset);
    pcbs[offset].processState = RUNNING;

     // Update the current executing process index.
    currentProcessIndex = offset;

    // Update the stack pointer to reflect the next item.
    stackp = pcbs[currentProcessIndex].stackPtr;
    writesp();
    load_context();
}






